#ifndef BUDDY_H
#define BUDDY_H

#include <stddef.h>

#define BUDDY_EXP_AUTOFREE    0x00000001u /* automatically free up memory */
#define BUDDY_EXP_ALLOCPOW2   0x00000002u /* allocate memory rounded up to power of 2 */
#define BUDDY_EXP_PREALLOC    0x00000004u /* preallocate space by overallocating */
#define BUDDY_EXP_INCREMENTAL 0x00000008u /* always allocate a larger memory block than the previous */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


/* opaque type of the buddy allocator state */
struct buddy;


/* Retrieve the supported expansion flags (BUDDY_EXP_*) */
unsigned int buddy_get_supported_expansion_flags(void);


/* Initialize the given memory block into a buddy allocator heap. */
/* All internal state will be allocated from the given memory block, */
/* decreasing the total allocatable memory. Since no extra memory */
/* allocation will be made, the whole heap can be destroyed by just */
/* freeing up the original memory block. */
/* This function may return NULL, if the heap does not fit in the */
/* given memory block. */
struct buddy* buddy_init(
	/* start address of the memory block (aligment will be applied */
	/* during the initialization) */
	void* memory,
	/* size of the memory block in bytes */
	size_t size,
	/* size of the smallest allocatable block */
	/* (there is a hard limit on this, which is 2*sizeof(void*), */
	/* but this limit will be automatically applied during the init; */
	/* must be power of two) */
	size_t minsize);


/* Similar to buddy_init(), but the initialized buddy allocator */
/* can be extended with new memory chunks. The returned shall be */
/* destroyed with buddy_deinit(). Apart from this, the allocator */
/* works the same. */
struct buddy* buddy_init_extendable(
	/* see buddy_init() */
	void* memory,
	/* see buddy_init() */
	size_t size,
	/* see buddy_init() */
	size_t minsize,
	/* number of desired expansion slots, including the first one */
	size_t max_expansions,
	/* flags (BUDDY_EXP_*) */
	unsigned int expansion_flags,
	/* context of the external allocator */
	void* ctx,
	/* function to be called when the allocator needs more memory */
	void* (*alloc_)(void*,size_t),
	/* function to be called to free unnecessary memory */
	void (*free_)(void*,void*));


/* Deinitializes the buddy allocator. This function has no effect */
/* on allocators created with buddy_init(), however it has important */
/* activitied on allocators created with buddy_init_extendable(). */
void buddy_deinit(
	struct buddy* const buddy);


/* Allocate a memory block. Blocks allocated with this function shall */
/* be freed up using buddy_free(). */
void* buddy_alloc(
	struct buddy* const buddy,
	size_t size);


/* Free a previously allocated memory block. */
void buddy_free(
	struct buddy* const buddy,
	/* address of a memory block that was returned by buddy_alloc() */
	void* block);


/* Get the size of the allocated memory block. */
/* Returns zero for addresses are currently not allocated. */
size_t buddy_msize(
	struct buddy* const buddy,
	/* address of a memory block that was returned by buddy_alloc() */
	void* block);


/* Get the size of the smallest allocatable block. */
/* This may be interesting when minimal size was near the hard limit. */
size_t buddy_get_smallest_size(
	struct buddy* const buddy);


/* Get the size of the theoretically largest allocatable block. */
/* This value is determined during the initialization. */
size_t buddy_get_largest_size(
	struct buddy* const buddy);


/* Get the total allocatable size of the heap. */
/* The returned value does not include the size of the internal state. */
size_t buddy_get_total_size(
	struct buddy* const buddy);


/* Get the size of the largest allocatable block. */
/* This may return zero, when no more free space is available. */
size_t buddy_get_largest_free(
	struct buddy* const buddy);


/* Populate the total free space. This may take some time. */
size_t buddy_get_total_free(
	struct buddy* const buddy);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* BUDDY_H */
